static __global__ void 
assign_mult_gamma5_kernel(      cuhqfield *gye,
                          const cuhqfield  *ye)
{
////////////////////////////////////////////////////////////////////////////////
// Multiply gamma5 : gye = gamma5 ye
//
//     gye : even/odd site fermion (ieo=0/1) (output)
//      ye : odd/even site fermion (ieo=0/1) (input)
//
////////////////////////////////////////////////////////////////////////////////
  __align__(16) float4 yri[CSH],gyri[CSH];

  yri[0] = (*ye).b[blockIdx.x].yri[0].c[threadIdx.x];
  yri[1] = (*ye).b[blockIdx.x].yri[1].c[threadIdx.x];
  yri[2] = (*ye).b[blockIdx.x].yri[2].c[threadIdx.x];
  yri[3] = (*ye).b[blockIdx.x].yri[3].c[threadIdx.x];
  yri[4] = (*ye).b[blockIdx.x].yri[4].c[threadIdx.x];
  yri[5] = (*ye).b[blockIdx.x].yri[5].c[threadIdx.x];

  __yr11(gyri) = __yr13(yri);
  __yr21(gyri) = __yr23(yri);
  __yr31(gyri) = __yr33(yri);
  __yr12(gyri) = __yr14(yri);
  __yr22(gyri) = __yr24(yri);
  __yr32(gyri) = __yr34(yri);
  __yr13(gyri) = __yr11(yri);
  __yr23(gyri) = __yr21(yri);
  __yr33(gyri) = __yr31(yri);
  __yr14(gyri) = __yr12(yri);
  __yr24(gyri) = __yr22(yri);
  __yr34(gyri) = __yr32(yri);

  __yi11(gyri) = __yi13(yri);
  __yi21(gyri) = __yi23(yri);
  __yi31(gyri) = __yi33(yri);
  __yi12(gyri) = __yi14(yri);
  __yi22(gyri) = __yi24(yri);
  __yi32(gyri) = __yi34(yri);
  __yi13(gyri) = __yi11(yri);
  __yi23(gyri) = __yi21(yri);
  __yi33(gyri) = __yi31(yri);
  __yi14(gyri) = __yi12(yri);
  __yi24(gyri) = __yi22(yri);
  __yi34(gyri) = __yi32(yri);

  (*gye).b[blockIdx.x].yri[0].c[threadIdx.x] = gyri[0];
  (*gye).b[blockIdx.x].yri[1].c[threadIdx.x] = gyri[1];
  (*gye).b[blockIdx.x].yri[2].c[threadIdx.x] = gyri[2];
  (*gye).b[blockIdx.x].yri[3].c[threadIdx.x] = gyri[3];
  (*gye).b[blockIdx.x].yri[4].c[threadIdx.x] = gyri[4];
  (*gye).b[blockIdx.x].yri[5].c[threadIdx.x] = gyri[5];
}

static __global__ void 
accum_mult_gamma5_kernel(cuhqfield *ye)
{
////////////////////////////////////////////////////////////////////////////////
// Multiply gamma5 : ye := gamma5 ye
//
//      ye : odd/even site fermion (ieo=0/1) (input/output)
//
////////////////////////////////////////////////////////////////////////////////

  __align__(16) float4 yri[CSH],gyri[CSH];

  yri[0] = (*ye).b[blockIdx.x].yri[0].c[threadIdx.x];
  yri[1] = (*ye).b[blockIdx.x].yri[1].c[threadIdx.x];
  yri[2] = (*ye).b[blockIdx.x].yri[2].c[threadIdx.x];
  yri[3] = (*ye).b[blockIdx.x].yri[3].c[threadIdx.x];
  yri[4] = (*ye).b[blockIdx.x].yri[4].c[threadIdx.x];
  yri[5] = (*ye).b[blockIdx.x].yri[5].c[threadIdx.x];

  __yr11(gyri) = __yr13(yri);
  __yr21(gyri) = __yr23(yri);
  __yr31(gyri) = __yr33(yri);
  __yr12(gyri) = __yr14(yri);
  __yr22(gyri) = __yr24(yri);
  __yr32(gyri) = __yr34(yri);
  __yr13(gyri) = __yr11(yri);
  __yr23(gyri) = __yr21(yri);
  __yr33(gyri) = __yr31(yri);
  __yr14(gyri) = __yr12(yri);
  __yr24(gyri) = __yr22(yri);
  __yr34(gyri) = __yr32(yri);

  __yi11(gyri) = __yi13(yri);
  __yi21(gyri) = __yi23(yri);
  __yi31(gyri) = __yi33(yri);
  __yi12(gyri) = __yi14(yri);
  __yi22(gyri) = __yi24(yri);
  __yi32(gyri) = __yi34(yri);
  __yi13(gyri) = __yi11(yri);
  __yi23(gyri) = __yi21(yri);
  __yi33(gyri) = __yi31(yri);
  __yi14(gyri) = __yi12(yri);
  __yi24(gyri) = __yi22(yri);
  __yi34(gyri) = __yi32(yri);

  (*ye).b[blockIdx.x].yri[0].c[threadIdx.x] = gyri[0];
  (*ye).b[blockIdx.x].yri[1].c[threadIdx.x] = gyri[1];
  (*ye).b[blockIdx.x].yri[2].c[threadIdx.x] = gyri[2];
  (*ye).b[blockIdx.x].yri[3].c[threadIdx.x] = gyri[3];
  (*ye).b[blockIdx.x].yri[4].c[threadIdx.x] = gyri[4];
  (*ye).b[blockIdx.x].yri[5].c[threadIdx.x] = gyri[5];
}

static void 
assign_mult_gamma5( cuhqfield *gye, const cuhqfield *ye)
{
  assign_mult_gamma5_kernel<<<BLOCKS,THREADS>>>(gye,ye);
  CUT_CHECK_ERROR("assign_mult_gamma5 kernel execution faild");
}

static void 
accum_mult_gamma5( cuhqfield *ye)
{
  accum_mult_gamma5_kernel<<<BLOCKS,THREADS>>>(ye);
  CUT_CHECK_ERROR("accum_mult_gamma5 kernel execution faild");
}
