#!/usr/bin/ruby
#
# Generates 4-D SSOR Kernel for Hermitian Squared Domainwall operator (D'D)
#
# $Id: GenMultSSOR.rb,v 1.2 2010/09/01 04:21:07 ishikawa Exp $
#

require 'FieldClasses'

state = SiteRestriction.new("","blocknormal")

amu = Array.new(5)
for i in 1..5
   amu[i-1] = i
end

puts "!================================================="

yt = Spinor.new("yt(iw)%y")
st = Spinor.new("st(iw)%y")
y  = Spinor.new("y")
w  = Spinor.new("w")

z  = FieldDWSpinor.new("z")
b  = FieldDWSpinor.new("q")
x  = FieldDWSpinor.new("dq")
vu = FieldGauge.new("vu","vector")

##############
# st = Hop z
##############
puts "  do iw=1,NS"
puts "    st(iw)%y(:,:) = Z0"
puts "  enddo"

##########################################
#  st = D'z
##########################################
for mu in amu.reverse
  puts  "!---------------------------"
  print "!  mu=",mu,"\n"
  puts  "!---------------------------"

  ##########################################
  #  st(iw) = st(iw) + (1+gamma_mu)U_mu(n) z(n+mu)
  ##########################################
puts "  do iw=1,NS"
  state.extra_boundary(+mu)
  if (mu != 5)
    puts  y.assign_proj_up(+mu,z.field_shift(+mu))
    puts  w.assign_mult_u_half(vu.dir(mu).field,y,+1)
    puts st.accum_proj_low(+mu,w)
  else
puts "    iwup = mod(iw-1+1+NS,NS)+1"
puts "    iwdn = mod(iw-1-1+NS,NS)+1"
    puts  y.assign_proj_up(+mu,z.field_shift(+mu))
    puts st.accum_proj_low(+mu,y)
  end
puts "  enddo"

  ##############################################
  #  st(iw) = st(iw) + (1-gamma_mu)U_mu(n-mu)^+ z(n-mu)
  ##############################################
puts "  do iw=1,NS"
  state.extra_boundary(-mu)
  if (mu != 5)
    puts  y.assign_proj_up(-mu,z.field_shift(-mu))
    puts  w.assign_mult_u_half(vu.dir(mu).field_shift(-mu),y,-1)
    puts st.accum_proj_low(-mu,w)
  else
puts "    iwup = mod(iw-1+1+NS,NS)+1"
puts "    iwdn = mod(iw-1-1+NS,NS)+1"
    puts  y.assign_proj_up(-mu,z.field_shift(-mu))
    puts st.accum_proj_low(-mu,y)
  end
puts "  enddo"

end

##################################################
# st(iw) = (z - kappa * st(iw)) = D'D x = D'z
##################################################
puts ""
puts "  do iw=1,NS"
puts "    st(iw)%y(:,:) = " + z.field.get_name + "(:,:) - kappa*st(iw)%y(:,:)"
puts "  enddo"
puts ""

#################################################
# multiply inverse diagnal part of (D'D)
#  yt = Diag \ st = Diag \ D' z
#################################################
print "
  do iw=1,NS
    xy(:,1,iw) = (st(iw)%y(:,1) - st(iw)%y(:,3))*0.5_DP
    xy(:,2,iw) = (st(iw)%y(:,2) - st(iw)%y(:,4))*0.5_DP
    xw(:,1,iw) = (st(iw)%y(:,1) + st(iw)%y(:,3))*0.5_DP
    xw(:,2,iw) = (st(iw)%y(:,2) + st(iw)%y(:,4))*0.5_DP
  enddo
  do iw=1,NS
    yy(:,:,iw) = Z0
    ww(:,:,iw) = Z0
    do jw=1,NS
      yy(:,1,iw) = yy(:,1,iw) + m_invXL(iw,jw)*xy(:,1,jw)
      yy(:,2,iw) = yy(:,2,iw) + m_invXL(iw,jw)*xy(:,2,jw)
      ww(:,1,iw) = ww(:,1,iw) + m_invXR(iw,jw)*xw(:,1,jw)
      ww(:,2,iw) = ww(:,2,iw) + m_invXR(iw,jw)*xw(:,2,jw)
    enddo
    yt(iw)%y(:,1) =  yy(:,1,iw) + ww(:,1,iw)
    yt(iw)%y(:,2) =  yy(:,2,iw) + ww(:,2,iw)
    yt(iw)%y(:,3) = -yy(:,1,iw) + ww(:,1,iw)
    yt(iw)%y(:,4) = -yy(:,2,iw) + ww(:,2,iw)
  enddo
\n"

#################################################
# overrelaxation
# st = b - yt*omega
#################################################
puts "  do iw=1,NS"
puts "    st(iw)%y(:,:) = " + b.field.get_name + "(:,:) - yt(iw)%y(:,:)*omega"
puts "  enddo"
puts ""

##################################################
# Update solution
# x(n) = st(n)
##################################################
puts "if ((ibx == 1 .or. ibx == NBX) .or. &"
puts "    (iby == 1 .or. iby == NBY) .or. &"
puts "    (ibz == 1 .or. ibz == NBZ)) then"
puts "!$OMP BARRIER"
puts "endif"

puts "  do iw=1,NS"
#puts "    " + x.field.get_name + "(:,:) = " + x.field.get_name + "(:,:) + st(iw)%y(:,:)"
puts "    " + x.field.get_name + "(:,:) = st(iw)%y(:,:)"
puts "  enddo"

##################################################
# Update auxiliary vector z
# z = D x (keep this relation)
##################################################

#######################
# z(n) = z(n) + st(n)
#######################

puts "  do iw=1,NS"
puts "    " + z.field.get_name + "(:,:) = " + z.field.get_name + "(:,:) + st(iw)%y(:,:)"
puts "  enddo"


puts ""
puts "  do iw=1,NS"
puts "    st(iw)%y(:,:) = -kappa*st(iw)%y(:,:)"
puts "  enddo"


for mu in amu.reverse

  ###############################################################
  # z(n-mu) = z(n-mu) + (1-gamma_mu)U_mu(n-mu) st(n)
  ###############################################################
  puts  "!---------------------------"
  print "!  mu=",mu,"\n"
  puts  "!---------------------------"
  puts ""
puts "  do iw=1,NS"
  state.extra_boundary(-mu)
  if (mu != 5)
    puts y.assign_proj_up(-mu,st)
    puts w.assign_mult_u_half(vu.dir(mu).field_shift(-mu),y,+1)
    puts z.field_shift(-mu).accum_proj_low(-mu,w)
  else
puts "    iwup = mod(iw-1+1+NS,NS)+1"
puts "    iwdn = mod(iw-1-1+NS,NS)+1"
    puts y.assign_proj_up(-mu,st)
    puts z.field_shift(-mu).accum_proj_low(-mu,y)
  end
puts "  enddo"


  ################################################################
  # z(n+mu) = z(n+mu) - kappa * (1+gamma_mu)U_mu(n)^dag st(n)
  ################################################################
puts "  do iw=1,NS"
  state.extra_boundary(+mu)
  if (mu != 5)
    puts y.assign_proj_up(+mu,st)
    puts w.assign_mult_u_half(vu.dir(mu).field,y,-1)
    puts z.field_shift(+mu).accum_proj_low(+mu,w)
  else
puts "    iwup = mod(iw-1+1+NS,NS)+1"
puts "    iwdn = mod(iw-1-1+NS,NS)+1"
    puts y.assign_proj_up(+mu,st)
    puts z.field_shift(+mu).accum_proj_low(+mu,y)
  end
puts "  enddo"
end
